# -*- coding: utf-8 -*-
import matplotlib.pyplot as plt
import matplotlib.image as mpimg
import matplotlib.colors as colors
from operator import attrgetter
import numpy as np
import weles
import operator

from HS import *
from HSFrame import *

"""
**HSSA** is a two-staged segmentation tool for hs images. It is based on consecutive division of image to decompose it into a set of homogenous segments. Later, according to labels assigned by an expert, the representation is prepared as an input for classification methods.
"""

class HSSA:
    """
    ## Initialization
    """
    def __init__(self, hs, threshold, jThreshold, limit=99, points=20):
        """
        Assign:

        - `hs` — hyperspectral image,
        - `threshold` — homogeneity threshold,
        - `limit` — iteration limiter,
        """
        self.hs = hs
        self.threshold = threshold
        self.jThreshold = jThreshold
        self.limit = limit
        self.points = points

        """
        And initialize:

        - `iteration` — number of current iteration,
        - `isComplete` — completion flag,
        - `segments` - segments counter
        """
        self.iteration = 0
        self.isComplete = False
        self.segments = 0
        self.classes = {}
        self.maxlabel = 0

        # Initialize sets of frames and set up representation.
        self.heterogenous = []
        self.homogenous = []
        self.clean()

    """
    ## Single iteration
    """
    def step(self):
        self.iteration += 1
        self.separate()

        for frame in self.homogenous:
            frame.setHomo()
        self.isComplete = len(self.heterogenous) == 0

        # Breaking hetero
        newHeterogenous = []
        for frame in self.heterogenous:
            newHeterogenous.extend(frame.divide())
        self.heterogenous = newHeterogenous

        # Merge procedure at every iteration
        self.merge()

    """
    ## Whole loop
    """
    def process(self):
        while not self.isComplete:
            if self.iteration == self.limit:
                break
            self.step()
        self.merge()

    """
    ## Merging process
    """
    def merge(self):
        for i in xrange(len(self.homogenous)):
            frame = self.homogenous[i]
            if frame.segment == -1:
                for j in xrange(i):
                    cmpFrame = self.homogenous[j]
                    similarity = \
                        1 - np.mean(np.std([frame.signature, cmpFrame.signature], axis=0))
                    if similarity > self.jThreshold:
                        frame.segment = cmpFrame.segment
                        break
                if frame.segment == -1:
                    frame.segment = self.segments
                    self.segments += 1

    """
    ## Algorithm output

    - 0 — one mean signature from mean of frames
    - 1 — one mean signature per frame
    - 2 - n random signatures from every frame
    """
    def train(self, c):
        train = []
        for label in self.classes:
            collection = self.classes[label]
            #print '- %i frames for label %i:' % (
            #    len(collection),
            #    label
            #)

            # One per class
            if c == 0:
                buffer = []
                for frame in collection:
                    signature = frame.signature
                    buffer.append(signature)
                    #for sample in samples:
                    #    print "\t- %s" % sample.features[:3]
                signature = np.mean(buffer, axis=0)
                sample = weles.Sample(
                    signature,
                    self.hs.reverseClasses[label]
                    )
                train.append(sample)
            # One per frame
            elif c == 1:
                #buffer = []
                for frame in collection:
                    #samples = frame.samples(3)
                    signature = frame.signature # one
                    sample = weles.Sample(
                        signature,
                        self.hs.reverseClasses[label]
                        )
                    train.append(sample)
                    #print "\tS=%s:" % signature[:3]
                    #buffer.append(signature)
                    #for sample in samples:
                    #    print "\t- %s" % sample.features[:3]
            else:
                #buffer = []
                for frame in collection:
                    signatures = frame.signatures(c)
                    signature = frame.signature # one
                    for signature in signatures:
                        sample = weles.Sample(
                            signature,
                            self.hs.reverseClasses[label]
                            )
                        train.append(sample)


        # print self.classes
        return train

    def representation(self):
        result = []
        for frame in self.homogenous:
            # print frame
            multiplier = self.iteration - frame.fold - 1
            units = pow(2, multiplier * 2)
            for signature in frame.signatures(units):
                line = [frame.label]
                line.extend(signature)
                result.append(line)
        return result

    ## Helper functions

    """
    ### Status string
    """
    def __str__(self):
        return "%s %s, i = %i, %i het, %i hom, s = %i" % (
            "complete" if self.isComplete else "in progress",
            self.hs.name,
            self.iteration,
            len(self.heterogenous),
            len(self.homogenous),
            self.segments
        )

    """
    ### Preparing representation to work
    """
    def clean(self):
        self.iteration = 0
        self.heterogenous = [HSFrame(self.hs, self.points)]
        self.homogenous = []
        self.isComplete = False

    """
    ### Generating png preview
    """
    def png(self, title = False, labels = False):
        # Join FLR-s
        union = self.heterogenous + self.homogenous

        # Generate title if not provided.
        if not title:
            title = 'hssa_i%i_t%.0f.png' % (
                self.iteration,
                1000 * self.threshold)

        # Establish base resolution on a iterated power of 2.
        base = pow(2, self.iteration)
        img = np.ones((base, base, 3))

        if len(union):
            # Scale intensivity according to values in FLR-s.
            minN = min(union, key=attrgetter('intensity')).intensity
            maxN = max(union, key=attrgetter('intensity')).intensity

            # Iterate every frame
            for frame in union:
                amount = pow(2, frame.fold)
                length = base / amount
                intensity = (frame.intensity - minN) / (maxN - minN)
                hue = 0
                if frame.isHomo:
                    if labels:
                        hue = frame.label / float(self.hs.maxlabel)
                    else:
                        hue = frame.segment / float(2 * len(self.hs.classes))
                        if self.segments > 2 * len(self.hs.classes):
                            hue = frame.segment / float(self.segments)
                x = frame.location % amount
                y = frame.location / amount
                for i in xrange(length):
                    for j in xrange(length):
                        if frame.isHomo:
                            if labels:
                                img[length * x + i, length * y + j] = \
                                    colors.hsv_to_rgb([hue, .75, .75])
                            else:
                                img[length * x + i, length * y + j] = \
                                    colors.hsv_to_rgb([hue, 1, .5 + intensity / 2])
                        else:
                            img[length * x + i, length * y + j] = \
                                colors.hsv_to_rgb([0, 0, intensity])

        # Plot
        plt.imshow(img, interpolation="nearest")
        plt.axis('off')
        plt.title('%s image, iteration %i, t = %.3f\n%i homo / %i hetero / %i segments' % (
            self.hs.name,
            self.iteration,
            self.threshold,
            len(self.homogenous),
            len(self.heterogenous),
            self.segments
        ))
        plt.savefig(title)

    """
    ### Separate frames
    """
    def separate(self):
        # separating
        #print 'Separating'
        #for frame in self.heterogenous:
        #    print frame
        self.homogenous.extend(
            [x for x in self.heterogenous if x.homogeneity > self.threshold])
        self.heterogenous = \
            [x for x in self.heterogenous if x.homogeneity <= self.threshold]

    """
    ### Postprocessing
    """
    def post(self):
        # Removing the background
        self.heterogenous = []
        self.homogenous = \
            [x for x in self.homogenous if x.label != 0]

        # Analyzing segment by segment
        for segment in xrange(self.segments):
            labels = []
            for frame in self.homogenous:
                if frame.segment == segment:
                    labels.append((frame.label, frame.fold, frame.homogeneity))

            if len(labels):
                votes = {}
                sizeOfSegment = 0
                for label in labels:
                    # print label
                    value = pow(4, self.iteration - label[1] - 1)
                    sizeOfSegment += value
                    value *= label[2]
                    if not label[0] in votes:
                        votes[label[0]] = value
                    else:
                        votes[label[0]] += value

                # Zmniejszenie roli tła
                #print votes
                if 0 in votes:
                    # print "%f is too large" % votes[0]
                    votes[0] /= 3
                #print votes

                label = max(votes.iteritems(), key=operator.itemgetter(1))[0]
                if label > self.maxlabel:
                    self.maxlabel = label

                #print "Segment %i of size %i labeled as %i" % (
                #    segment, sizeOfSegment, label)
                #print 'Votes: %s gives us decision %i' % (votes, label)
                for frame in self.homogenous:
                    if frame.segment == segment:
                        frame.label = label
                        if not label in self.classes:
                            self.classes[label] = [frame]
                        else:
                            self.classes[label].append(frame)

        # Removing the background
        self.heterogenous = []
        self.homogenous = \
            [x for x in self.homogenous if x.label != 0]

        # print analysis
        print "%i classes detected" % len(self.classes)



# -*- coding: utf-8 -*-
import scipy.io
import numpy as np
import weles
import operator

"""
A class to store and to use hypersectral image.
"""
class HS:
    """
    ## Initialization
    """
    def __init__(self, dictionary):
        # Loading image, ground truth and establishing informations.
        self.image = self.loadMatFromTuple(dictionary['image'])
        self.gt = self.loadMatFromTuple(dictionary['gt'])
        self.name = dictionary['name']
        self.classes = dictionary['classes']

        #print type(self.image[(0,0)][0])

        # Getting in shape.
        shape = np.shape(self.image)
        self.rows = shape[0]
        self.cols = shape[1]
        self.bands = shape[2]

        # Normalize
        self.normA = []
        self.normB = []
        for sliceIdx in xrange(self.bands):
            slice = self.slice(sliceIdx)
            #print sliceIdx
            minimum = np.min(slice)
            maximum = np.max(slice)
            self.normA.append(minimum)
            self.normB.append(maximum)
            #print slice
            #print 'min %i max %i' % (minimum, maximum)
        self.normB = map(operator.sub, self.normB, self.normA)
        #print self.normA
        #print self.normB


        # Searching for maximum value
        self.max = np.amax(self.image)
        self.maxlabel = np.amax(self.gt)

        self.reverseClasses = None
        self.prepareReverse()

    ## Operators

    """
    ### Getting sample
    """
    def sample(self, location):
        return weles.Sample(
            self.signature(location),
            self.label(location))

    """
    ### Getting signature
    """
    def signature(self, location):
        return (self.image[location].astype(float) - self.normA) / self.normB

    """
    ### Getting label
    """
    def label(self, location):
        return int(np.copy(self.gt[location]))

    """
    ### Getting slice
    """
    def slice(self, band):
        return np.copy(self.image[:, :, band])

    ## Helper functions

    """
    ## Loading from .mat file
    """
    def loadMatFromTuple(self, entry):
        return scipy.io.loadmat(entry[0])[entry[1]]

    def prepareReverse(self):
        self.reverseClasses = {}
        for (x,y), value in np.ndenumerate(self.gt):
            label = self.label((x, y))
            if not label in self.reverseClasses:
                self.reverseClasses.update({label: len(self.reverseClasses)})

    """
    ## Exporting Weles dataset
    """
    def dataset(self):
        dataset = weles.Dataset()
        db_name = self.name
        source_samples = []
        self.reverseClasses = {}
        for x in xrange(self.rows):
            for y in xrange(self.cols):
                sample = self.sample((x, y))
                source_samples.append(sample)
                if not sample.label in self.reverseClasses:
                    self.reverseClasses.update({sample.label: len(self.reverseClasses)})
                sample.label = self.reverseClasses[sample.label]
        dataset.fill(
            db_name,
            source_samples,
            self.reverseClasses
        )
        print '# Classes\n%s' % self.reverseClasses
        return dataset


    """
    ## Be verbose, man
    """
    def __str__(self):
        return '%s image, %i classes, %i samples of %i bands' % (
            self.name,
            len(self.classes),
            self.rows * self.cols,
            self.bands)



import scipy.io
import numpy as np
import random
import weles
import math

SEED = 123
random.seed(SEED)

class Window:
    def __init__(self, frame):
        # Getting base
        base = pow(2, frame.fold)

        # Calculating size
        width = float(frame.hs.cols) / base
        height = float(frame.hs.rows) / base

        # Establishing position
        x = int(frame.location / base)
        y = int(frame.location % base)
        self.top = int(width * x)
        self.left = int(height * y)
        self.width = int(width)
        self.height = int(height)

        if self.width == 0:
            self.width = 1

        if self.height == 0:
            self.height = 1
        '''
        print '[C%i R%i L%i X%i Y%i] t:%i l:%i w:%i(%.2f) h:%i(%.2f)' %(
            frame.hs.cols, frame.hs.rows,
            frame.location, x, y,
            self.top, self.left,
            self.width, width,
            self.height, height
        )
        '''
    def __str__(self):
        return "T:%i L:%i W:%i H:%i" % (
            self.top, self.left, self.width, self.height
        )

class HSFrame:
    def __init__(self, hs, points=20, fold=0, location=0):
        self.points = points  # amount of points to create mean signature
        self.hs = hs  # image
        self.fold = fold            # iteration, when frame was created
        self.location = location  # iteration based frame location
        self.segment = -1  # segment identifier
        self.label = -1  # class label given by the expert for region
        self.homogeneity = 0  # homogeneity measure
        self.isHomo = False
        self.intensity = 0  # intensity measure
        self.signature = []  # mean frame signature

        # Fill
        self.window = Window(self)
        self.calculate()

    def setHomo(self):
        self.isHomo = True

    def divide(self):
        # Bases
        base = pow(2, self.fold)
        newBase = pow(2, self.fold + 1)

        # Counting overlines
        overlines = self.location / base

        # Establishing start
        x = 2 * (self.location + self.location / base * base)

        # Creating new frames
        frames = []
        frames.append(
            HSFrame(self.hs, self.points, self.fold + 1, x))
        frames.append(
            HSFrame(self.hs, self.points, self.fold + 1, x + 1))
        frames.append(
            HSFrame(self.hs, self.points, self.fold + 1, x + newBase))
        frames.append(
            HSFrame(self.hs, self.points, self.fold + 1, x + newBase + 1))
        return frames

    def signatures(self, amount):
        width = self.window.width
        height = self.window.height
        top = self.window.top
        left = self.window.left
        stop = height * width
        signatures = []
        for item in xrange(0, amount if amount < stop else stop):
            index = random.randrange(stop)
            x = int(index / width)
            y = int(index % width)
            signature = self.hs.signature((left + x, top + y))
            signatures.append(signature)
        return signatures

    def samples(self, amount):
        width = self.window.width
        height = self.window.height
        top = self.window.top
        left = self.window.left
        stop = height * width
        samples = []
        for item in xrange(0, amount):
            index = random.randrange(stop)
            x = int(index / width)
            y = int(index % width)
            sample = self.hs.sample((left + x, top + y))
            samples.append(sample)
        return samples

    def calculate(self):
        #print '### Calculate'
        # Getting window parameters
        width = self.window.width
        height = self.window.height
        top = self.window.top
        left = self.window.left

        # Calculating index range
        stop = height * width
        points = self.points
        if self.points > stop:
            points = stop

        # Getting signatures
        signatures = []
        labels = []
        for item in xrange(0, points):
            index = random.randrange(stop)
            x = int(index / width)
            y = int(index % width)
            signature = self.hs.signature((left + x, top + y))
            signatures.append(signature)
            labels.append(self.hs.label((left + x, top + y)))
            #print '## ITEM %s' % signature[:3]

        self.signature = np.mean(signatures, axis=0)
        #print '## SIGN %s' % self.signature[:3]

        self.label = max(set(labels), key=labels.count)
        # homogeneity as a mean standardDeviation
        self.homogeneity = \
            1 - np.mean(np.std(signatures, axis=0))
        #print '## HOMO %f' % self.homogeneity
        #print '%f' % (1 - np.mean(np.std(signatures, axis=0)))

        self.intensity = np.mean(self.signature)

    def __str__(self):
        return "F%i|L%i|S%i|L%i|H%.3f" % (
            self.fold,
            self.location,
            self.segment,
            self.label,
            self.homogeneity)
